import sys
import fractions
import json
import uuid
import random
import string
from typing import Sequence

import arrow
from datetime import datetime, timezone

from .bases import Field, MultiField


if sys.version_info < (3, 7):
    def check_iso(value):
        return arrow.get(value).isoformat()
else:
    def check_iso(value):
        return datetime.fromisoformat(value).replace(tzinfo=timezone.utc).isoformat()


class Any(Field):
    def cast(self, value):
        return value

    def sample(self):
        return 4


class Boolean(Field):
    def cast(self, value):
        if isinstance(value, str):
            if value[0:5].lower() == 'false':
                return False
        return bool(value)

    def sample(self):
        return random.getrandbits(1) == 0


class Integer(Field):
    def cast(self, value):
        return int(value)

    def sample(self):
        return random.randint(-2**30, 2**30)


class Float(Field):
    def cast(self, value):
        return float(value)

    def sample(self):
        return random.random() * 1000 - 500


class SeparatedFraction(MultiField):
    """Store a fraction as two values.

    Most
    """
    def cast(self, value) -> Sequence:
        if isinstance(value, fractions.Fraction):
            pass
        elif isinstance(value, (tuple, list)):
            value = fractions.Fraction(*value)
        else:
            value = fractions.Fraction(value)
        return value.numerator, value.denominator

    def proxy(self, parent, values):
        fraction = fractions.Fraction(*values)
        parent[self.name + '_numerator'] = fraction.numerator
        parent[self.name + '_denominator'] = fraction.denominator
        return fraction

    def sample(self):
        return fractions.Fraction(random.randint(-1000000, 100000), random.randint(0, 100000))

    def components(self):
        return (
            self.name + '_numerator',
            self.name + '_denominator'
        )


class String(Field):
    def cast(self, value):
        if isinstance(value, bytes):
            return value.decode()
        return str(value)

    def sample(self):
        length = random.randint(0, 128)
        return ''.join(random.choices(string.ascii_letters + string.digits + string.whitespace, k=length))


class Bytes(Field):
    def cast(self, value):
        if isinstance(value, str):
            return value.encode()
        return bytes(value)

    def sample(self):
        length = random.randint(0, 2**18)
        return bytes(random.getrandbits(8) for _ in range(length))


class Keyword(String):
    """A short string with symbolic value."""
    def sample(self):
        length = random.randint(0, 128)
        return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


class UUID(Keyword):
    def __init__(self, **kwargs):
        if kwargs.get('factory') == 'random':
            kwargs['factory'] = lambda: uuid.uuid4().hex
        super().__init__(**kwargs)


class Text(String):
    """A string with natural content."""
    def sample(self):
        chunks = random.randint(0, 128)
        return '\n'.join(super(Text, self).sample() for _ in range(chunks))


class Timestamp(Float):
    """A floating point number representing an offset from epoch"""
    pass


class DateString(String):
    """A field storing date."""
    def cast(self, value):
        if value == "NOW":
            return datetime.utcnow().isoformat()

        try:
            return check_iso(value)
        except (TypeError, ValueError):
            return arrow.get(value).isoformat()

    def sample(self):
        return '2020-03-20T14:28:23.382748'


class Enum(Field):
    """A field for enum values."""
    def __init__(self, enum, **kwargs):
        super().__init__(**kwargs)
        self.enum = enum
        self.conversion = {}
        for val in self.enum:
            self.conversion[val.value] = val
            self.conversion[val.name] = val
            self.conversion[val] = val

    def sample(self):
        return random.choice(list(self.conversion.values()))

    def cast(self, value):
        try:
            return self.conversion[value]
        except (KeyError, TypeError):
            raise ValueError(f"Not an accepted enum value {value}")


class JSON(String):
    """A string field that checks that its content is always valid JSON"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def cast(self, value):
        value = super().cast(value)
        json.loads(value)
        return value

    def sample(self):
        return random.choice([
            '{}',
            '[]',
            '0',
            '"abc"',
            'null',
        ])



from .basic import Boolean, Integer, Float, Timestamp, Enum, Bytes
from .basic import String, Keyword, Text, JSON, UUID, DateString, SeparatedFraction
from .basic import Any
from .pattern import PatternString, MD5, SHA256, PhoneNumber, MACAddress, PrivateIP, SSDeepHash, IP, \
    Domain, Email, SHA1, URI, URIPath
from .complex import Compound


def List(field_type, **kwargs):
    """A helper "class" to choose the right list class for a given field."""
    from .bases import MultivaluedField, ProxyField
    from .multivalued import SimpleList
    from .complex import List
    if isinstance(field_type, ProxyField):
        return List(field_type, **kwargs)
    return SimpleList(field_type, **kwargs)


def Mapping(field_type, **kwargs):
    """A helper "class" to choose the right list class for a given field."""
    from .bases import MultivaluedField, ProxyField
    from .multivalued import SimpleMapping
    from .complex import Mapping
    if isinstance(field_type, ProxyField):
        return Mapping(field_type, **kwargs)
    return SimpleMapping(field_type, **kwargs)



from typing import Dict, Tuple, Any, Sequence, Optional


class Field:
    """An abstract data field for a model."""
    def __init__(self, **kwargs):
        self.metadata: Dict[str, Any] = kwargs
        self.metadata_defaults: Dict[str, Any] = {}
        self.name: Optional[str] = None

    def __contains__(self, item):
        return item in self.metadata or item in self.metadata_defaults

    def __getitem__(self, item):
        return self.metadata.get(item, self.metadata_defaults.get(item))

    def cast(self, value):
        raise NotImplementedError()

    def sample(self):
        """Generate a random value appropriate for this field."""
        raise NotImplementedError(self.__class__)


class MultivaluedField(Field):
    """A base for fields where the value could have many parts, but don't require a proxy."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def cast(self, value):
        raise NotImplementedError()

    def sample(self):
        raise NotImplementedError()

    def flat_fields(self, prefix):
        raise NotImplementedError()


class ProxyField(Field):
    """A base for fields where the underlying data and the 'view' are different."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def cast(self, value) -> Tuple[Any, Any]:
        """Convert value into the type for this field.

        The cast method for proxy is special in that it returns two things:
            - The proxy object that code interacts with
            - The underlying data object that will occupy the field in the parent dictionary.

        The proxy is responsible for making sure what it presents is kept up to
        date with whatever is returned from this method for data. For example, the list
        and mapping proxy objects return proxies that wrap a dictionary or list object
        that are returned as the second parameter.

        Proxy fields need to have matching names, and any complex data has to be
        contained within the field itself (this is opposed to MultiFields where the
        opposite is true)
        """
        raise NotImplementedError()

    def sample(self):
        raise NotImplementedError()

    def flat_fields(self, prefix):
        raise NotImplementedError()


class MultiField(Field):
    """A field that actually represents several hidden entries in the model.

    Uses a proxy similar to a proxy field, but doesn't have its own internal
    data structure, instead relying on storing data in the parent structure.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def cast(self, value) -> Sequence:
        """Convert an assigned value into a sequence of values suitable to pass to proxy."""
        raise NotImplementedError()

    def proxy(self, parent, values):
        """Construct a proxy over the values given.

        The values will which will match in order the keys produced by 'components'.

        Either this method or the proxy object is responsible for setting the
        values it is constructed with into the parent in the corresponding fields.

        If this method does it, the proxy should be something immutable, if its not
        immutable, then the proxy is responsible for retaining a reference to parent
        and keeping the parent synced with the view of it.
        """
        raise NotImplementedError()

    def sample(self):
        """Return an (ideally random) object that would be appropriate as an argument to cast."""
        raise NotImplementedError()

    def components(self):
        """Return the keys for the components.

        This must be the same every time this is called for a given field.
        """
        raise NotImplementedError()



""""""
import json
import weakref

import typing
from typing import Dict, Set

from .fields.bases import ProxyField, Field, MultiField

_fields: Dict[type, Dict[str, Field]] = typing.cast(Dict, weakref.WeakKeyDictionary())
_flat_fields: Dict[type, Dict[str, Field]] = typing.cast(Dict, weakref.WeakKeyDictionary())


def model_fields(cls: type):
    return _fields[cls]


def model_fields_flat(cls: type):
    return _flat_fields[cls]


def raw(obj):
    return getattr(obj, '_data')


def dumps(obj):
    return json.dumps(raw(obj))


def model(cls=None, **metadata):
    # If we are given default metadata
    if cls is None:
        def capture(cls):
            return model(cls, **metadata)
        return capture

    # Track the keys that will be added to the model so that we can
    # check if two fields conflict in what keys they use (primarily
    # that multi-fields don't try and use the same hidden keys)
    keys: Set[str] = set()

    fields = {}        # The fields of the object
    flat_fields = {}   # The fields of the object, recursively flattened
    compounds = {}     # Compound type fields
    multi_fields = {}  # MultiValue type fields
    basic = {}         # Fields with simple types only
    casts = {}         # Each field's cast function
    properties = {}    # Any previously defined properties
    methods = {}       # Any methods from the class we want to preserve
    static_values = {}

    for _name, field in cls.__dict__.items():
        if isinstance(field, Field):
            casts[_name] = field.cast
            fields[_name] = field
            field.name = _name
            field.metadata_defaults = metadata

            if field.metadata.get('optional', False):
                def make_optional_cast(_c):
                    def _cast(value):
                        if value is None:
                            return None
                        return _c(value)
                    return _cast
                field.cast = casts[_name] = make_optional_cast(field.cast)

            if isinstance(field, ProxyField):
                keys.add(_name)
                compounds[_name] = field
                for sub_name, sub_field in field.flat_fields(prefix=_name).items():
                    flat_fields[sub_name] = sub_field
            elif isinstance(field, MultiField):
                multi_fields[_name] = field
            else:
                keys.add(_name)
                basic[_name] = field
                flat_fields[_name] = field

        elif isinstance(field, property):
            keys.add(_name)
            properties[_name] = field

        elif callable(field) or isinstance(field, (classmethod, staticmethod)):
            keys.add(_name)
            methods[_name] = field

        elif not _name.startswith('__') and not _name.endswith('__'):
            keys.add(_name)
            static_values[_name] = field

    field_names = set(fields.keys())

    # Now go back over the multi fields, and make sure none of their generated
    # sub components collide with anything else in the object we are creating.
    multi_field_components = {}
    proxies = {}
    for _name, field in multi_fields.items():
        _x = multi_field_components[_name] = tuple(field.components())
        for component in _x:
            if component in keys:
                raise ValueError(f"Error creating model {cls.__name__} collision on key {component} with field {_name}")
            else:
                keys.add(component)

                # Add the component names to the field_names list so they are legal
                # arguments to construct/exist already in an object being checked
                # against the model
                field_names.add(component)
        proxies[_name] = field.proxy

        # If the multi field hasn't already used its name, reserve it.
        keys.add(_name)

    def field_property(_name, _cast, optional):
        if optional:
            class FieldProperty:
                def __get__(self, instance, objtype):
                    return instance._data.get(_name)

                def __set__(self, instance, value):
                    instance._data[_name] = _cast(value)

        else:
            class FieldProperty:
                def __get__(self, instance, objtype):
                    return instance._data[_name]

                def __set__(self, instance, value):
                    instance._data[_name] = _cast(value)

        return FieldProperty()

    class CompoundProperty:
        def __init__(self, name, field):
            self.name = name
            self.field = field

        def __get__(self, instance, objtype):
            return instance._compounds[self.name]

        def __set__(self, instance, value):
            instance._compounds[self.name], instance._data[self.name] = casts[self.name](value)

    class MultiValueProperty:
        def __init__(self, name, field):
            self.name = name
            self.field = field

        def __get__(self, instance, objtype):
            return instance._compounds[self.name]

        def __set__(self, instance, value):
            instance._compounds[self.name] = proxies[self.name](instance._data, casts[self.name](value))

    class ModelClass:
        __slots__ = ['_data', '_compounds']

        def __init__(self, *args, **kwargs):
            data = self._data = args[0] if args else {}
            _compounds = self._compounds = {}

            if not isinstance(data, dict):
                raise ValueError("Unexpected parameter type for model construction")

            kw_pop = kwargs.pop

            if set(data.keys()) - field_names:
                raise ValueError(f"Unexpected key provided: {set(data.keys()) - field_names}")

            for name, field in compounds.items():
                if name in kwargs:
                    _compounds[name], data[name] = field.cast(kw_pop(name))
                elif data.get(name) is not None:
                    _compounds[name], data[name] = field.cast(data[name])
                elif 'default' in field.metadata:
                    _compounds[name], data[name] = field.cast(field['default'])
                elif 'factory' in field.metadata:
                    _compounds[name], data[name] = field.cast(field['factory']())
                elif field.metadata.get('optional', False):
                    _compounds[name] = None
                    # data[name] = None
                    continue
                else:
                    raise ValueError(f"Missing key [{name}] to construct {cls.__name__}")

            for name, field in multi_fields.items():
                _components = multi_field_components[name]
                if all(_c in kwargs for _c in _components):
                    _compounds[name] = field.proxy(data, [kw_pop(_c) for _c in _components])
                elif name in kwargs:
                    _compounds[name] = field.proxy(data, field.cast(kw_pop(name)))
                elif all(_c in data for _c in _components):
                    _compounds[name] = field.proxy(data, [data[_c] for _c in _components])
                elif data.get(name) is not None:
                    _compounds[name] = field.proxy(data, field.cast(data[name]))
                elif 'default' in field.metadata:
                    _compounds[name] = field.proxy(data, field.cast(field['default']))
                elif 'factory' in field.metadata:
                    _compounds[name] = field.proxy(data, field.cast(field['factory']()))
                elif field.metadata.get('optional', False):
                    _compounds[name] = None
                    # data[name] = None
                    continue
                else:
                    raise ValueError(f"Missing key [{name}] to construct {cls.__name__}")

            for name, field in basic.items():
                cast = casts[name]
                if name in kwargs:
                    data[name] = cast(kw_pop(name))
                elif name in data:
                    data[name] = cast(data[name])
                elif 'default' in field.metadata:
                    data[name] = cast(field['default'])
                elif 'factory' in field.metadata:
                    data[name] = cast(field['factory']())
                elif field.metadata.get('optional', False):
                    pass
                    # data[name] = None
                else:
                    raise ValueError(f"Missing key [{name}] to construct {cls.__name__}")

            if kwargs:
                raise ValueError(f"Unexpected key provided: {kwargs.keys()}")

        def __eq__(self, other):
            if isinstance(other, dict):
                return self == self.__class__(**other)
            for name in fields.keys():
                print(name, getattr(self, name), getattr(other, name))
                if getattr(self, name) != getattr(other, name):
                    print('false')
                    return False
                print('true')
            return True

    # Lets over write some class properties to make it a little nicer
    ModelClass.__name__ = cls.__name__
    ModelClass.__doc__ = cls.__doc__
    if hasattr(cls, '__annotations__'):
        ModelClass.__annotations__ = cls.__annotations__

    _fields[ModelClass] = fields
    _flat_fields[ModelClass] = flat_fields

    # Apply the properties to the class so that our attribute access works
    for _name, field in compounds.items():
        setattr(ModelClass, _name, CompoundProperty(_name, field))
    for _name, field in basic.items():
        setattr(ModelClass, _name, field_property(_name, field.cast, field['optional']))
    for _name, field in multi_fields.items():
        setattr(ModelClass, _name, MultiValueProperty(_name, field.cast))

    # If there were any pre-defined properties on the class make sure it is put back
    for _name, _p in properties.items():
        setattr(ModelClass, _name, _p)
    for _name, _p in methods.items():
        setattr(ModelClass, _name, _p)
    for _name, _p in static_values.items():
        setattr(ModelClass, _name, _p)

    return ModelClass



import os
from setuptools import setup, find_packages
from pkg_resources import parse_version

modules = None
try:
    from Cython.Build import cythonize
    modules = cythonize([
        'draughts/*.py',
        'draughts/*/*.py'
    ], language_level=3)
    print("Using cython")
except ImportError:
    cythonize = None
    print("Using pure python")


version_string = os.environ.get('DRAUGHTS_VERSION', os.environ.get('GITHUB_REF', "0.0.0").lstrip('refs/tags/v'))
version_string = str(parse_version(version_string))
print("Building version ", version_string)

try:
    readme = open(os.path.join(os.path.dirname(__file__), 'readme.md')).read()
except FileNotFoundError:
    readme = None

setup(
    name="draughts",
    version=version_string,
    packages=find_packages(exclude='test'),
    ext_modules=modules,
    
    # metadata to display on PyPI
    author="Adam Douglass",
    author_email="douglass@malloc.ca",
    description="Generates boilerplate for data objects.",
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords="utility typechecking",
    url="https://github.com/adam-douglass/draughts/",
    install_requires=[
        'rstr',
        'arrow'
    ],
    extras_require={
        'test': ['pytest', 'pytest-subtests'],
        'speedup': ['cython']
    }
)
